// FileStream.cpp
//------------------------------------------------------------------------------

// Includes
//------------------------------------------------------------------------------
#include "FileStream.h"

#include <Core/Env/Assert.h>
#include <Core/Strings/AString.h>
#include <Core/Strings/AStackString.h>

#include <stdio.h>

// Defines
//------------------------------------------------------------------------------
//#define FILESTREAM_DETAILED_DEBUG // uncomment this for lots of spam
#ifdef FILESTREAM_DETAILED_DEBUG
	#include <Core/Tracing/Tracing.h>
	#include <Core/Process/Thread.h>
	#include <windows.h>
#endif

// CONSTRUCTOR
//------------------------------------------------------------------------------
FileStream::FileStream()
: m_Handle( nullptr )
{
}

// DESTRUCTOR
//------------------------------------------------------------------------------
FileStream::~FileStream()
{
	if ( IsOpen() )
	{
		Close();
	}
}

// Open
//------------------------------------------------------------------------------
bool FileStream::Open( const char * fileName, uint32_t fileMode )
{
	ASSERT( !IsOpen() );

	// access mode
	AStackString<8> options;
	if ( ( fileMode & READ_ONLY ) != 0 )
	{
		ASSERT( fileMode == READ_ONLY ); // no extra flags allowed
		options += 'r';
	}
	else if ( ( fileMode & WRITE_ONLY ) != 0 )
	{
		options += 'w';
	}
	else
	{
		ASSERT( false ); // must specify an access mode
	}

	options += 'b'; // always in binary mode

	// extra flags
	if ( ( fileMode & TEMP ) != 0 )
	{
		options += 'T'; // windows = don't flush to disc if possible
	}
	if ( ( fileMode & DELETE_ON_CLOSE ) != 0 )
	{
		options += 'D'; // windows - delete when closed
	}

	if ( fopen_s( (FILE **)&m_Handle, fileName, options.Get() ) != 0 )
	{
		#ifdef FILESTREAM_DETAILED_DEBUG
			DEBUGSPAM( "Open file error: %d - '%s' on %x\n", GetLastError(), fileName, Thread::GetCurrentThreadId() );
		#endif
		m_Handle = nullptr;
		return false;
	}
	#ifdef FILESTREAM_DETAILED_DEBUG
		DEBUGSPAM( "Open file OK: %x - '%s' on %x\n", m_Handle, fileName, Thread::GetCurrentThreadId() );
	#endif

	ASSERT( m_Handle );
	return true;
}

// Close
//------------------------------------------------------------------------------
void FileStream::Close()
{
	ASSERT( IsOpen() );
	VERIFY( fclose( (FILE *)m_Handle ) == 0 );
	#ifdef FILESTREAM_DETAILED_DEBUG
		DEBUGSPAM( "Close file: %x on %x\n", m_Handle, Thread::GetCurrentThreadId() );
	#endif
	m_Handle = nullptr;
}

// Read
//------------------------------------------------------------------------------
/*virtual*/ uint64_t FileStream::ReadBuffer( void * buffer, uint64_t bytesToRead )
{
	ASSERT( buffer );
	ASSERT( IsOpen() );

	uint64_t totalBytesRead = 0;
	do
	{
		uint64_t remaining = ( bytesToRead - totalBytesRead );
		size_t tryToReadNow = ( remaining > MEGABYTE ) ? MEGABYTE : (size_t)remaining;
		uint64_t bytesReadNow = fread( (char *)buffer + (size_t)totalBytesRead, 1, tryToReadNow, (FILE *)m_Handle );
		if ( bytesReadNow == 0 )
		{
			break; // failed
		}
		totalBytesRead += bytesReadNow;
	} while ( totalBytesRead < bytesToRead );

	#ifdef FILESTREAM_DETAILED_DEBUG
		DEBUGSPAM( "Read %llu bytes to %x on %x\n", totalBytesRead, m_Handle, Thread::GetCurrentThreadId() );
	#endif

	return totalBytesRead;
}

// Write
//------------------------------------------------------------------------------
/*virtual*/ uint64_t FileStream::WriteBuffer( const void * buffer, uint64_t bytesToWrite )
{
	ASSERT( buffer );
	ASSERT( IsOpen() );

	uint64_t totalBytesWritten = 0;
	do
	{
		uint64_t remaining = ( bytesToWrite - totalBytesWritten );
		size_t tryToWriteNow = ( remaining > MEGABYTE ) ? MEGABYTE : (size_t)remaining;
		uint64_t bytesWrittenNow = fwrite( (char *)buffer + (size_t)totalBytesWritten, 1, tryToWriteNow, (FILE *)m_Handle );
		if ( bytesWrittenNow == 0 )
		{
			break; // failed
		}
		totalBytesWritten += bytesWrittenNow;
	} while ( totalBytesWritten < bytesToWrite );

	#ifdef FILESTREAM_DETAILED_DEBUG
		DEBUGSPAM( "Write %llu bytes to %x on %x\n", totalBytesWritten, m_Handle, Thread::GetCurrentThreadId() );
	#endif

	return totalBytesWritten;
}

// Flush
//------------------------------------------------------------------------------
/*virtual*/ void FileStream::Flush()
{
	#ifdef FILESTREAM_DETAILED_DEBUG
		OUTPUT( "Flush %x on %x\n", m_Handle, Thread::GetCurrentThreadId() );
	#endif

	ASSERT( IsOpen() );
	VERIFY( fflush( (FILE *)m_Handle ) == 0 );
}

// Tell
//------------------------------------------------------------------------------
/*virtual*/ uint64_t FileStream::Tell() const
{
	#ifdef FILESTREAM_DETAILED_DEBUG
		OUTPUT( "Tell %x on %x\n", m_Handle, Thread::GetCurrentThreadId() );
	#endif

	ASSERT( IsOpen() );

	uint64_t pos = _ftelli64( (FILE *)m_Handle );
	return pos;
}

// Seek
//------------------------------------------------------------------------------
/*virtual*/ bool FileStream::Seek( uint64_t pos ) const
{
	#ifdef FILESTREAM_DETAILED_DEBUG
		OUTPUT( "Seek to %llu, %x on %x\n", pos, m_Handle, Thread::GetCurrentThreadId() );
	#endif

	ASSERT( IsOpen() );

	int retVal = _fseeki64( (FILE *)m_Handle, pos, SEEK_SET );
	return ( retVal == 0 );
}

// GetFileSize
//------------------------------------------------------------------------------
/*virtual*/ uint64_t FileStream::GetFileSize() const
{
	#ifdef FILESTREAM_DETAILED_DEBUG
		OUTPUT( "GetFileSize %x on %x\n", m_Handle, Thread::GetCurrentThreadId() );
	#endif

	ASSERT( IsOpen() );

	// record current pos to restore
	uint64_t originalPos = Tell();

	// seek to the end
	VERIFY( _fseeki64( (FILE *)m_Handle, 0, SEEK_END ) == 0 );

	// get the current pos now that we're at the ends
	uint64_t size = _ftelli64( (FILE *)m_Handle );

	// seek back to the original pos
	VERIFY( Seek( originalPos ) );

	return size;
}

//------------------------------------------------------------------------------
